// Copyright (C) 2024 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#include "ds.h"

#include <QJsonDocument>
#include <QJsonObject>
#include <QScreen>

DesignStudio::DesignStudio(QWebSocket *socket, const QString &m_deviceUuid, QObject *parent)
    : QObject(parent)
    , m_deviceUuid(m_deviceUuid)
    , m_socket(socket)
{
    initPingPong();
    initSocket();

    m_pingTimer.start();
}

void DesignStudio::initPingPong()
{
    m_pingTimer.setInterval(1000);
    m_pongTimer.setInterval(30000);
    m_pongTimer.setSingleShot(true);
    m_pingTimer.setSingleShot(true);

    connect(&m_pingTimer, &QTimer::timeout, this, [this]() {
        m_socket->ping();
        m_pongTimer.start();
    });

    connect(m_socket.data(),
            &QWebSocket::pong,
            this,
            [this](quint64 elapsedTime, const QByteArray &) {
                m_pongTimer.stop();
                m_pingTimer.start();
            });

    connect(&m_pongTimer, &QTimer::timeout, this, [this]() {
        qDebug() << "Design Studio" << m_id << "is not responding. Closing the connection.";
        m_socket->close();
        m_socket->abort();
        emit disconnected(m_id);
    });
}

void DesignStudio::initSocket()
{
    connect(m_socket.data(), &QWebSocket::disconnected, this, [this]() {
        qDebug() << "Design Studio" << m_id << "disconnected";
        m_pingTimer.stop();
        m_pongTimer.stop();
        emit disconnected(m_id);
    });

    connect(m_socket.data(),
            &QWebSocket::textMessageReceived,
            this,
            &DesignStudio::processTextMessage);

    connect(m_socket.data(),
            &QWebSocket::binaryFrameReceived,
            this,
            [this](const QByteArray &frame, bool isLastFrame) {
                qDebug() << "Binary frame received. Size:" << frame.size()
                         << "is last: " << isLastFrame;
                m_projectData.append(frame);
                emit projectIncomingProgress(m_id,
                                             m_projectData.size() * 100 / m_incomingProjectSize);
                if (isLastFrame) {
                    emit projectReceived(m_id, m_projectData);
                }
            });
}

QString DesignStudio::ipv4Addr() const
{
    return m_socket->peerAddress().toString();
}

QString DesignStudio::id() const
{
    return m_id;
}

void DesignStudio::sendDeviceInfo()
{
    const QRect screenGeometry = QGuiApplication::primaryScreen()->geometry();

    QJsonObject deviceInfo;
    deviceInfo["screenHeight"] = screenGeometry.height();
    deviceInfo["screenWidth"] = screenGeometry.width();
    deviceInfo["os"] = QSysInfo::prettyProductName();
    deviceInfo["osVersion"] = QSysInfo::productVersion();
    deviceInfo["architecture"] = QSysInfo::currentCpuArchitecture();
    deviceInfo["deviceId"] = m_deviceUuid;
    deviceInfo["appVersion"] = QString(CMAKE_VAR_GIT_VERSION);

    qDebug() << "Sending device info to Design Studio" << deviceInfo;

    sendData(PackageToDesignStudio::deviceInfo, deviceInfo);
}

void DesignStudio::sendData(const QLatin1String &dataType, const QJsonValue &data)
{
    QJsonObject message;
    message["dataType"] = dataType;
    message["data"] = data;

    m_socket->sendTextMessage(QJsonDocument(message).toJson(QJsonDocument::Compact));
}

void DesignStudio::processTextMessage(const QString &message)
{
    QJsonParseError jsonError;
    const QJsonDocument jsonDoc = QJsonDocument::fromJson(message.toLatin1(), &jsonError);

    if (jsonError.error != QJsonParseError::NoError) {
        qDebug() << "Failed to parse JSON message:" << jsonError.errorString() << message;
        return;
    }

    const QJsonObject jsonObj = jsonDoc.object();
    if (!jsonObj.contains("dataType")) {
        qDebug() << "Invalid JSON message:" << jsonObj;
        return;
    }

    const QString dataType = jsonObj.value("dataType").toString();

    if (dataType == PackageFromDesignStudio::designStudioReady) {
        const QString newDesignStudioId = jsonObj.value("data").toString();
        qDebug() << "Design Studio ready with ID" << newDesignStudioId;
        m_id = newDesignStudioId;
        sendDeviceInfo();
        emit idReceived(newDesignStudioId, ipv4Addr());
    } else if (dataType == PackageFromDesignStudio::projectData) {
        m_incomingProjectSize = jsonObj.value("data").toInt();
        qDebug() << "Project is expected with size" << m_incomingProjectSize;
        m_projectData.clear();
        emit projectIncoming(m_incomingProjectSize);
    } else if (dataType == PackageFromDesignStudio::stopRunningProject) {
        qDebug() << "Stop running project requested";
        emit projectStopRequested(m_id);
    } else {
        qDebug() << "Unkown JSON message";
    }
}

// this function should only be triggered when the device is receiving a qmlrc project
void DesignStudio::processBinaryMessage(const QByteArray &data)
{
    qDebug() << "Binary message received (most probably a project)";
    emit projectReceived(m_id, data);
}

void DesignStudio::sendProjectRunning()
{
    sendData(PackageToDesignStudio::projectRunning);
}

void DesignStudio::sendProjectStopped()
{
    sendData(PackageToDesignStudio::projectStopped);
}

void DesignStudio::sendProjectLogs(const QString &logs)
{
    sendData(PackageToDesignStudio::projectLogs, logs);
}
